/*  BMP - Cross-platform multimedia player
 *  Copyright (C) 2003-2004  Edward Brocklesby, Chong Kai Xiong,
 *            Milosz Derezynski, David Lau, Michiel Sikkes
 *
 *  ui.c/bmp_ui.h
 *  bmp_button.c/bmp_button.h
 *  bmp_button_toggle.c/bmp_button_toggle.h
 *  bmp_slider.c/bmp_slider.h
 *
 *  M.Derezynski
 *  Tim-Philipp Mueller
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <gtk/gtk.h>

#include <main.h>

#include <bmp_ui.h>
#include <ui_playlist.h>
#include <ui_main.h>
#include <ui_util.h>
#include <ui_callbacks.h>
#include <ui_prefswin.h>

#include <error.h>
#include <util.h>
#include <bmp_resource_manager.h>

#include <bmp_button.h>
#include <bmp_button_toggle.h>
#include <bmp_slider.h>
#include <bmp_window.h>
#include <chroma_list.h>

G_DEFINE_TYPE (BmpUI, bmp_ui, G_TYPE_OBJECT)

/* GObject stuff */
enum {
    BMP_UI_SIGNAL_0,

    BMP_UI_N_SIGNALS
};

#if 0
static guint signals[BMP_UI_N_SIGNALS] = { 0 };
#endif

enum {
    BMP_PLAY_N_PROPS
};

struct _BmpUIPrivate {

    gboolean dispose_has_run;

    BmpWindowMain         *window_main;
    BmpWindowPlaylist     *window_playlist;
    BmpWindowPreferences  *window_preferences;

    GdkBitmap             *window_main_mask;
};

static gchar* bmp_skin_files[] = {
	"main.bmp",
	"shufrep.bmp",
	"cbuttons.bmp",
	"posbar.bmp",
	"volume.bmp",
	"titlebar.bmp",
	"balance.bmp",
	"pledit.bmp",
	"monoster.bmp",
	NULL,
	"playpaus.bmp"
};

static BmpSkinComponent bmp_skin_files_sequential[] = {
	SKIN_COMPONENT_MAIN,
	SKIN_COMPONENT_SHUFREP,
	SKIN_COMPONENT_CBUTTONS,
	SKIN_COMPONENT_TITLEBAR,
	SKIN_COMPONENT_PLEDIT,
	SKIN_COMPONENT_MONOSTER,
	SKIN_COMPONENT_PLAYPAUSE
};

/* 'extern' referenced */
GtkUIManager *bmp_ui_manager = NULL;
BmpChromaWidget *chroma_widgets[BMP_N_CHROMA_WIDGETS];

/* Action groups */
GtkActionGroup *bmp_actions = NULL;
GtkActionGroup *bmp_toggle_actions = NULL;

static void
load_skin_components(BmpUI *self, const gchar * archive_filename, gint type) {

  GdkPixbuf *pixbuf;
  const gchar *archive_dirname;
  gchar *filename;
  gint n;

 /* Free memory */
  for (n = 0; n < N_SKIN_COMPONENTS; n++)
  {
     if (bmp_skin_component[n] != NULL)
     {
	g_object_unref (bmp_skin_component[n]->pixbuf);
	g_free (bmp_skin_component[n]);
	bmp_skin_component[n] = NULL;
     }
  }

  if (bmp_posbar)
	{
		g_object_unref (bmp_posbar->pixbuf);
		g_free (bmp_posbar);
	}

  if (bmp_volume)
	{
		g_object_unref (bmp_volume->pixbuf);
		g_free (bmp_volume);
	}

  if (bmp_balance)
	{
		g_object_unref (bmp_balance->pixbuf);
		g_free (bmp_balance);
	}

  /* LOAD COMPONENTS */
  if (type == BMP_SKIN_ARCHIVE) {
  	archive_dirname = archive_decompress(archive_filename);
  } else {
  	archive_dirname = archive_filename;
  };

  bmp_cursors_free();
  bmp_cursors_set(archive_dirname);

  /* Buttons, Titlebars */
  for (n = 0; n < G_N_ELEMENTS(bmp_skin_files_sequential); n++)
  {
       filename = find_file_recursively(archive_dirname, bmp_skin_files[bmp_skin_files_sequential[n]]);
      if (! filename) {
        	g_message("%s not found", bmp_skin_files[bmp_skin_files_sequential[n]]);
      }
      else
      {
	      pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	      bmp_skin_component[bmp_skin_files_sequential[n]] = g_malloc0(sizeof(bmp_skin_component_t));
	      bmp_skin_component[bmp_skin_files_sequential[n]]->identifier = bmp_skin_files_sequential[n];
	      bmp_skin_component[bmp_skin_files_sequential[n]]->pixbuf = pixbuf;
	      g_free(filename);
      }
  }

  /* Numbers */
  bmp_skin_component[SKIN_COMPONENT_NUMBERS] = g_malloc0(sizeof(bmp_skin_component_t));
  filename = find_file_recursively(archive_dirname, "numbers.bmp");
  if (! filename)
  {
	filename = find_file_recursively(archive_dirname, "nums_ex.bmp");
	if (! filename)
		g_message("%s: Numbers not found", G_STRFUNC);
  }

  if  (filename != NULL)
  {
	pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	bmp_skin_component[SKIN_COMPONENT_NUMBERS]->identifier = SKIN_COMPONENT_NUMBERS;
	bmp_skin_component[SKIN_COMPONENT_NUMBERS]->pixbuf = pixbuf;

	if (gdk_pixbuf_get_width(pixbuf) == 108)
	    bmp_skin_component[SKIN_COMPONENT_NUMBERS]->user_data = GINT_TO_POINTER (12);
	else
	    bmp_skin_component[SKIN_COMPONENT_NUMBERS]->user_data = GINT_TO_POINTER (10);

	g_free(filename);
  }
  else
  {
	g_free(bmp_skin_component[SKIN_COMPONENT_NUMBERS]);
  }

  /* Volume, Balance and Seek slider*/
  filename = find_file_recursively(archive_dirname, bmp_skin_files[SKIN_COMPONENT_POSBAR]);
  if (! filename) {
	g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_POSBAR]);
  }
  pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
  bmp_posbar = g_malloc0(sizeof(bmp_posbar_t));
  bmp_posbar->pixbuf = pixbuf;
  g_free(filename);

  filename = find_file_recursively(archive_dirname, bmp_skin_files[SKIN_COMPONENT_VOLUME]);
  if (! filename) {
	g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_VOLUME]);
  }
  pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
  bmp_volume = g_malloc0(sizeof(bmp_volume_t));
  bmp_volume->pixbuf = pixbuf;
  bmp_volume->height = gdk_pixbuf_get_height(pixbuf) - 422;

  if (gdk_pixbuf_get_height(pixbuf) > 422) {
  	bmp_volume->has_button = TRUE;
  } else {
  	bmp_volume->has_button = FALSE;
  };
  g_free(filename);

  filename = find_file_recursively(archive_dirname, bmp_skin_files[SKIN_COMPONENT_BALANCE]);
  if (! filename) {
	g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_BALANCE]);
  }
  pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
  bmp_balance = g_malloc0(sizeof(bmp_volume_t));
  bmp_balance->pixbuf = pixbuf;
  bmp_balance->height = gdk_pixbuf_get_height(pixbuf) - 421;
  if (gdk_pixbuf_get_height(pixbuf) > 421) {
  	bmp_balance->has_button = TRUE;
  } else {
  	bmp_balance->has_button = FALSE;
  };
  g_free(filename);

  bmp_colors_init(archive_dirname);

  /* FIXME: Move this into ui_main.c */
  self->private->window_main_mask = bmp_get_mask_window_main (archive_dirname);

  if (type == BMP_SKIN_ARCHIVE) {
  	del_directory(archive_dirname);
  };
}

GdkBitmap*
bmp_ui_get_mask_window_main (BmpUI *self)
{
    g_return_val_if_fail (BMP_IS_UI (self), NULL);

    return self->private->window_main_mask;
}

void
bmp_ui_display_windows(BmpUI *self) {

    gint x,y,w,h;
    gboolean visible;

    x = bmp_configuration_key_fetch_int (bmp_configuration, "main-window", "pos_x");
    y = bmp_configuration_key_fetch_int (bmp_configuration, "main-window", "pos_y");
    if ((x > -1) && (y > -1))
    {
	gtk_window_move (GTK_WINDOW(self->private->window_main->window), x, y);
	BMP_WINDOW (self->private->window_main->window)->x = x;
	BMP_WINDOW (self->private->window_main->window)->y = y;
    }
    else
    {
	x = gdk_screen_get_width (gdk_screen_get_default());
	y = gdk_screen_get_height (gdk_screen_get_default());

	x /= 2;
	y /= 2;

	x -= BMP_WINDOW_MAIN_WIDTH/2;
	y -= BMP_WINDOW_MAIN_HEIGHT/2;

	gtk_window_move (GTK_WINDOW(self->private->window_main->window), x, y);
	BMP_WINDOW (self->private->window_main->window)->x = x;
	BMP_WINDOW (self->private->window_main->window)->y = y;
    }
    gtk_widget_show_all (BMP_WINDOW_MAIN(self->private->window_main)->window);

    x = bmp_configuration_key_fetch_int (bmp_configuration, "playlist-window", "pos_x");
    y = bmp_configuration_key_fetch_int (bmp_configuration, "playlist-window", "pos_y");

    if ((x > -1) && (y > -1))
    {
	gtk_window_move (GTK_WINDOW(self->private->window_playlist->window), x, y);
	BMP_WINDOW (self->private->window_playlist->window)->x = x;
	BMP_WINDOW (self->private->window_playlist->window)->y = y;
    }
    else
    {
	x = gdk_screen_get_width (gdk_screen_get_default());
	y = gdk_screen_get_height (gdk_screen_get_default());

	x /= 2;
	y /= 2;

	x -= BMP_WINDOW_MAIN_WIDTH;
	y -= BMP_WINDOW_MAIN_HEIGHT;

	gtk_window_move (GTK_WINDOW(self->private->window_playlist->window), x, y);
	BMP_WINDOW (self->private->window_playlist->window)->x = x;
	BMP_WINDOW (self->private->window_playlist->window)->y = y;
    }

    w = bmp_configuration_key_fetch_int (bmp_configuration, "playlist-window", "width");
    h = bmp_configuration_key_fetch_int (bmp_configuration, "playlist-window", "height");

    if ((w > -1) && (h > -1))
	gtk_widget_set_size_request (GTK_WIDGET(self->private->window_playlist->window), w, h);
    else
	gtk_widget_set_size_request (GTK_WIDGET(self->private->window_playlist->window), BMP_WINDOW_MAIN_WIDTH*2, BMP_WINDOW_MAIN_HEIGHT*4);

    visible = bmp_configuration_key_fetch_boolean (bmp_configuration,"playlist-window","visible");

    if (visible) {
	bmp_system_control_show_pl_window (bmp_system_control, TRUE);
    }
}

static gint
load_skin(BmpUI *self, const gchar * filename)
{
    gint error_status;

    if (!g_file_test(filename, G_FILE_TEST_EXISTS)) {
	error_status = BMP_ERROR_FAILURE;
	return error_status;
    }

    bmp_colors_free();

    if (g_file_test(filename, G_FILE_TEST_IS_DIR)) {
        load_skin_components(self, filename, BMP_SKIN_DIR);
	error_status = BMP_ERROR_SUCCESS;
    }
    else
    if (g_file_test(filename, G_FILE_TEST_IS_REGULAR)) {
        load_skin_components(self, filename, BMP_SKIN_ARCHIVE);
	error_status = BMP_ERROR_SUCCESS;
    }
    else
    {
        error_status = BMP_ERROR_FAILURE;
	return error_status;
    }

    if (self->private->window_main)
        bmp_window_main_configure (BMP_WINDOW_MAIN (self->private->window_main));

    if (self->private->window_playlist)
	bmp_window_playlist_configure (BMP_WINDOW_PLAYLIST (self->private->window_playlist));

    return error_status;
}

static void
on_bmp_skin_changed (GObject *configuration, gpointer data)
{
	gchar *filename;

	filename = bmp_configuration_key_fetch (bmp_configuration, "bmp", "skin");
	if (!load_skin (BMP_UI(data), filename)) {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Unable to load skin '%s'", filename);
	}
}

static void
font_dispose (BmpFontInformation *fi)
{
    pango_font_description_free (fi->font_description);
    g_free (fi);
}

static void
on_bmp_font_changed (BmpUI *self)
{
    PangoFontDescription *font_desc;
    gchar *font;

    LOOKUP_OR_CREATE_DEFAULT ("bmp", "font", BMP_DEFAULT_FONT);
    font = bmp_configuration_key_fetch (bmp_configuration, "bmp", "font");

    if (bmp_fonts[BMP_FONT_NORMAL] != NULL)
	font_dispose (bmp_fonts[BMP_FONT_NORMAL]);
    bmp_fonts[BMP_FONT_NORMAL] =
	bmp_get_font_information (font);

    font_desc =
	pango_font_description_copy (bmp_fonts[BMP_FONT_NORMAL]->font_description);

    pango_font_description_set_size (font_desc,
	pango_font_description_get_size(font_desc)+ (1.2*PANGO_SCALE));

    if (bmp_fonts[BMP_FONT_LARGE] != NULL)
	font_dispose (bmp_fonts[BMP_FONT_LARGE]);
    bmp_fonts[BMP_FONT_LARGE]  =
	bmp_get_font_information_pfd (font_desc);
}

static gint
ui_initialize(BmpUI *self)
{
  gint rv, n;
  gchar *filename, *listname;

  if ((self->private->window_playlist != NULL) && (self->private->window_main != NULL)) return BMP_ERROR_SUCCESS;

  LOOKUP_OR_CREATE_DEFAULT ("bmp","skin", DATA_DIR G_DIR_SEPARATOR_S BMP_SKIN_DIR_BASENAME G_DIR_SEPARATOR_S BMP_DEFAULT_SKIN);
  filename =
	bmp_configuration_key_fetch(bmp_configuration, "bmp", "skin");

  for (n = 0; n < N_SKIN_COMPONENTS; bmp_skin_component[n++] = NULL);

  rv =
    load_skin (self, filename);

  if (rv != BMP_ERROR_SUCCESS) {
	g_free (filename);
	filename =
	  DATA_DIR G_DIR_SEPARATOR_S BMP_SKIN_DIR_BASENAME G_DIR_SEPARATOR_S BMP_DEFAULT_SKIN;
	bmp_configuration_key_set_string (bmp_configuration, "bmp","skin", filename);
	rv =
	    load_skin (self, filename);
  }

  if (rv == BMP_ERROR_SUCCESS) {
    gint n;

    g_object_connect (G_OBJECT(bmp_configuration),
			"signal::changed::bmp-skin",
			G_CALLBACK(on_bmp_skin_changed),
			self,
			"swapped-signal::changed::bmp-font",
			G_CALLBACK(on_bmp_font_changed),
			self,
			NULL);

    for (n = 0; n < BMP_N_FONTS; bmp_fonts[n++] = NULL);

    bmp_ui_manager = gtk_ui_manager_new();

    bmp_resource_manager_create_path (bmp_rm, NULL, "windows");
    bmp_resource_manager_create_path (bmp_rm, NULL, "tooltips");

    bmp_actions = gtk_action_group_new(BMP_MAIN_ACTION_GROUP);
    bmp_toggle_actions = gtk_action_group_new(BMP_MAIN_ACTION_GROUP_TOGGLE);

    gtk_action_group_set_translate_func (bmp_actions,
		bmp_menu_translate,
		NULL,
		NULL);
    gtk_action_group_set_translate_func (bmp_toggle_actions,
		bmp_menu_translate,
		NULL,
		NULL);

    self->private->window_main = bmp_window_main_new();
    RM_ITEM_NEW("windows", "window-main", self->private->window_main, NULL);

    g_signal_emit_by_name (G_OBJECT(bmp_configuration), "changed::bmp-font");

    self->private->window_playlist = bmp_window_playlist_new();
    RM_ITEM_NEW("windows", "window-playlist", self->private->window_playlist, NULL);

    self->private->window_preferences = bmp_window_preferences_new();
    RM_ITEM_NEW("windows", "window-preferences", self->private->window_preferences, NULL);

    bmp_window_main_configure (self->private->window_main);
    bmp_window_playlist_configure (self->private->window_playlist);

    bmp_ui_display_windows (self);

    chroma_list_set_list (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget), bmp_playlist_get_playlists(bmp_playlist));

    listname =
	bmp_configuration_key_fetch(bmp_configuration,"bmp","current-playlist");
    chroma_list_select_name (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget), listname);
    g_free (listname);

    LOOKUP_OR_CREATE_DEFAULT_BOOLEAN("bmp", "display-tracklist-numbers", TRUE);
    LOOKUP_OR_CREATE_DEFAULT_BOOLEAN("bmp", "shuffle", FALSE);
    LOOKUP_OR_CREATE_DEFAULT_BOOLEAN("bmp", "repeat", FALSE);
    bmp_system_control_repeat (bmp_system_control, bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "repeat"));
    bmp_system_control_shuffle (bmp_system_control, bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "shuffle"));

    chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));
  }

  return rv;
}

/* GObject stuff */
static void
bmp_ui_dispose (GObject *gobj)
{
    BmpUI *self = BMP_UI (gobj);

    if (self->private->dispose_has_run)
     {
         /*
          * if dispose did already run, return.
          */
         return;
     }
    /*
     * make sure dispose does not run twice.
     */
    self->private->dispose_has_run = TRUE;

    if (self->private->window_main) {

	gtk_widget_destroy(BMP_WINDOW(self->private->window_main->window)->canvas);
	gtk_widget_destroy(BMP_WINDOW(self->private->window_playlist->window)->canvas);
	BMP_WINDOW(self->private->window_main->window)->canvas = NULL;
	BMP_WINDOW(self->private->window_playlist->window)->canvas = NULL;

	gtk_widget_destroy(self->private->window_main->window);
	gtk_widget_destroy(self->private->window_playlist->window);

	g_object_unref (self->private->window_main);
	g_object_unref (self->private->window_playlist);
	g_object_unref (self->private->window_preferences);
	g_object_unref (bmp_ui_manager);

	bmp_ui_manager = NULL;
	self->private->window_main = NULL;
	self->private->window_playlist = NULL;
	self->private->window_preferences = NULL;
    }

    g_message("%s:\t\t\tSTOPPED", __FILE__);
}

static void
bmp_ui_class_init (BmpUIClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

#if 0
    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
    gobject_class->finalize = bmp_play_finalize;
    gobject_class->constructor = bmp_play_constructor;
#endif

    gobject_class->dispose = bmp_ui_dispose;
}

static void
bmp_ui_init (BmpUI *self)
{
    self->private = g_new0 (BmpUIPrivate, 1);

    self->private->window_main = NULL;
    self->private->window_playlist = NULL;
    self->private->window_preferences = NULL;

    self->private->window_main_mask = NULL;
}

BmpUI*
bmp_ui_new (void)
{
    BmpUI *self;

    self = g_object_new (BMP_TYPE_UI, NULL);
    return self;
}

gint
bmp_ui_finalize_init (BmpUI *self)
{
    gint rv;

    rv = ui_initialize(self);

    if (rv == BMP_ERROR_SUCCESS)
	    g_message("%s:\t\t\tINITIALIZED", __FILE__);
    else
	    g_message("%s:\t\t\tERROR: Not initalized: Skin couldn't be loaded", __FILE__);

    return rv;
}


