/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 */

#include <config.h>

#include <gtk/gtk.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <stdlib.h>

#include <errno.h>
#include <sys/stat.h>

#include <main.h>
#include <error.h>
#include <revision.h>

#include <bmp_sm.h>
#include <bmp_ui.h>
#include <bmp_playlist.h>
#include <bmp_configuration.h>
#include <bmp_play.h>
#include <bmp_resource_manager.h>
#include <bmp_metadata.h>
#include <fam_monitor.h>
#include <list_manager.h>
#include <logger.h>
#include <signals.h>

#ifdef HAVE_GST
#include <gst/gst.h>
#endif

#include <g_object_get.h>
#include <util.h>

gchar *bmp_paths[BMP_N_PATHS] = { NULL };

BmpResourceManager *bmp_rm = NULL;
BmpConfiguration   *bmp_configuration = NULL;
BmpPlaylist        *bmp_playlist = NULL;
BmpPlay		   *bmp_play = NULL;
BmpSystemControl   *bmp_system_control = NULL;
FamMonitor	   *bmp_fam_monitor = NULL;
BmpUI              *bmp_ui = NULL;
BmpMetadataCache   *bmp_metadata_cache = NULL;

static gchar *session_id = NULL;

static gboolean show_version  = FALSE;
static gboolean run_headless  = FALSE;
static gboolean become_daemon = FALSE;
static gboolean log_stdout    = FALSE;

#ifdef HAVE_DBUS
static gboolean no_remote     = FALSE;
#endif

static GOptionEntry bmp_options[] =
  {
    {"version", 'v', 0, G_OPTION_ARG_NONE, &show_version,
     N_("Display version"), NULL},
    {"no-log", 0, 0, G_OPTION_ARG_NONE, &log_stdout,
     N_("Disable logfile, log messages to shell"), NULL},
    {"no-ui", 0, 0, G_OPTION_ARG_NONE, &run_headless,
     N_("Disable UI"), NULL},
#ifdef HAVE_DBUS
    {"no-remote", 0, 0, G_OPTION_ARG_NONE, &no_remote,
     N_("Disable DBUS remote interface"), NULL},
#endif
    {"daemon", 0, 0, G_OPTION_ARG_NONE, &become_daemon,
     N_("Run as a daemon (implies --no-ui)"), NULL},
#ifdef HAVE_SM
    {"sm-client-id", 0, 0, G_OPTION_ARG_STRING, &session_id,
     N_("Specify session management ID"), N_("ID")},
#endif
    { NULL }
  };

static GList *
get_feature_list (void)
{
  GList *list = NULL;

#ifdef HAVE_SM
  list = g_list_append (list, N_("feature|X11 Session Management"));
#endif
#ifdef HAVE_DBUS
  list = g_list_append (list, N_("feature|D-BUS Support"));
#endif
#ifdef HAVE_XINE
  list = g_list_append (list, N_("feature|Audio backend: XINE"));
#endif
#ifdef HAVE_GST
  list = g_list_append (list, N_("feature|Audio backend: GStreamer"));
#endif

  return list;
}

static void
print_version (void)
{
  GList *features, *iter;

  g_print ("%s %s, Build: %s-R%s [%s] (%s)\n\n", PACKAGE, VERSION, RV_LAST_CHANGED_DATE, RV_REVISION, RV_DATE, RV_BUILDUSER);

  features = get_feature_list ();

  g_print (ngettext ("Feature built: \n", "Features built: \n",
                     g_list_length (features)));

  for (iter = features; iter; iter = g_list_next (iter))
    g_print ("* %s\n", Q_(iter->data));

  g_list_free (features);
}

static gboolean
make_user_dir (const gchar *name)
{
  const mode_t mode755 = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
  return make_directory (name, mode755);
}

static void
bmp_make_user_dirs (void)
{
  make_user_dir (bmp_paths[BMP_PATH_USER_DIR]);
  make_user_dir (bmp_paths[BMP_PATH_USER_SKIN_DIR]);
  make_user_dir (bmp_paths[BMP_PATH_SKIN_THUMB_DIR]);
  make_user_dir (bmp_paths[BMP_PATH_PLAYLISTS]);
}

static void
bmp_free_paths (void)
{
  int i;

  for (i = 0; i < BMP_N_PATHS; i++)
    {
      g_free (bmp_paths[i]);
      bmp_paths[i] = NULL;
    }
}

#define USER_PATH(path) \
    g_build_filename(bmp_paths[BMP_PATH_USER_DIR], path, NULL);

static void
bmp_init_paths (void)
{
  bmp_paths[BMP_PATH_USER_DIR] = g_build_filename(g_get_home_dir(), BMP_RCPATH, NULL);
  bmp_paths[BMP_PATH_USER_SKIN_DIR] = USER_PATH(BMP_SKIN_DIR_BASENAME);
  bmp_paths[BMP_PATH_SKIN_THUMB_DIR] = USER_PATH(BMP_SKIN_THUMB_DIR_BASENAME);
  bmp_paths[BMP_PATH_PLAYLISTS] = USER_PATH(BMP_PLAYLISTS_DIR_BASENAME);
  bmp_paths[BMP_PATH_LOG_FILE] = USER_PATH(BMP_LOG_BASENAME);
  bmp_paths[BMP_PATH_METADATA_CACHE] = USER_PATH(BMP_METADATA_CACHE_BASENAME);
#if 0
  bmp_paths[BMP_PATH_CONFIG_FILE] = USER_PATH(BMP_CONFIG_BASENAME);
  bmp_paths[BMP_PATH_ACCEL_FILE] = USER_PATH(BMP_ACCEL_BASENAME);
#endif

  bmp_make_user_dirs ();
}

static void
setup_i18n (void)
{
  gtk_set_locale ();
  bindtextdomain (PACKAGE, LOCALEDIR);
  bind_textdomain_codeset (PACKAGE, "UTF-8");
  textdomain (PACKAGE);
}

gint
main (gint    argc,
      gchar** argv)
{
  setup_i18n ();

  gtk_init_with_args (&argc, &argv, "", bmp_options, NULL, NULL);

  if (show_version)
    {
      print_version ();
      return EXIT_SUCCESS;
    }

#ifdef HAVE_GST
  gst_init (&argc, &argv);
#else
  g_thread_init (NULL);
#endif

  bmp_init_paths ();

  if (!log_stdout)
    bmp_logger_start (bmp_paths[BMP_PATH_LOG_FILE]);

  /* FIXME: gtk_init_with_args() (above) will remove all recognized
     arguments, so the session manager will run BMPx with incomplete
     options when it restarts. */
  session_start (argc, argv, session_id);

  bmp_rm = bmp_resource_manager_new ();
  bmp_configuration = bmp_configuration_new ();

  /* A small exception so we get the sequence right */
  LOOKUP_OR_CREATE_DEFAULT ("bmp", "playlists-path", bmp_paths[BMP_PATH_PLAYLISTS]);

  if (become_daemon)
    {
      if (!bmp_become_daemon ())
        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Unable to daemonize, proceeding with normal startup.");
      else
        run_headless = TRUE;
    }

  bmp_fam_monitor = fam_monitor_new ();

#ifdef HAVE_DBUS
  if (bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "no-remote"))
    no_remote = TRUE;
#endif

  bmp_playlist = bmp_playlist_new ();
  bmp_play = bmp_play_new ();

#ifdef HAVE_DBUS
  bmp_system_control = bmp_system_control_new (no_remote);
#else
  bmp_system_control = bmp_system_control_new (FALSE);
#endif

  if (bmp_system_control)
    {
      /* Note: the SIGTERM handler only works with a GTK main loop
       * - descender */
      signal_handlers_init ();

      bmp_metadata_cache = bmp_metadata_cache_new ();

      bmp_playlist_load_playlists (bmp_playlist);

      if (bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "no-ui"))
        run_headless = TRUE;

      if (!run_headless)
        bmp_system_control_ui_start (bmp_system_control);
      else
        g_message ("%s: Not starting UI on request", G_STRFUNC);

      gtk_main ();
    }

  if (bmp_play)
    bmp_play_stop_stream (bmp_play);

  if (bmp_metadata_cache)
    {
      g_object_unref (bmp_metadata_cache);
      bmp_metadata_cache = NULL;
    }

  if (bmp_play)
    {
      g_object_unref (bmp_play);
      bmp_play = NULL;
    }

  if (bmp_playlist)
    {
      g_object_unref (bmp_playlist);
      bmp_playlist = NULL;
    }

  if (bmp_fam_monitor)
    {
      g_object_unref (bmp_fam_monitor);
      bmp_fam_monitor = NULL;
    }

  if (bmp_configuration)
    {
      g_object_unref (bmp_configuration);
      bmp_configuration = NULL;
    }

  if (bmp_system_control)
    {
      g_object_unref (bmp_system_control);
      bmp_system_control = NULL;
    }

  session_end ();

  bmp_logger_stop ();

  bmp_free_paths ();

  exit (EXIT_SUCCESS);
}
